@@ -42,15 +42,7 @@
 extern crate cfg_if;
 
 cfg_if! {
-    if #[cfg(feature = "std")] {
-        mod imports {
-            pub use std::boxed;
-            pub use std::vec;
-            pub use std::string;
-            pub use std::borrow;
-            pub use std::collections::btree_map;
-        }
-    } else if #[cfg(feature = "alloc")] {
+    if #[cfg(all(not(feature = "std"), feature = "alloc"))] {
         extern crate core as std;
         #[macro_use]
         extern crate alloc;
@@ -61,6 +53,14 @@ cfg_if! {
             pub use alloc::borrow;
             pub use alloc::collections::btree_map;
         }
+    } else {
+        mod imports {
+            pub use std::boxed;
+            pub use std::vec;
+            pub use std::string;
+            pub use std::borrow;
+            pub use std::collections::btree_map;
+        }
     }
 }
 
@@ -210,9 +210,52 @@ substitutions = {:#?}",
 
         Ok(out)
     }
+
+    /// Demangle the symbol to a DemangleWrite, which lets the consumer be informed about
+    /// syntactic structure.
+    pub fn structured_demangle<W: DemangleWrite>(&self, out: &mut W, options: &DemangleOptions) -> fmt::Result {
+        let mut ctx = ast::DemangleContext::new(
+            &self.substitutions,
+            self.raw.as_ref(),
+            options,
+            out,
+        );
+        self.parsed.demangle(&mut ctx, None)
+    }
 }
 
-impl<T> Symbol<T> {
+/// The type of a demangled AST node.
+/// This is only partial, not all nodes are represented.
+#[derive(Clone, Copy, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub enum DemangleNodeType {
+    /// Entering a <prefix> production
+    Prefix,
+    /// Entering a <template-prefix> production
+    TemplatePrefix,
+    /// Entering a <template-args> production
+    TemplateArgs,
+}
+
+/// Sink for demangled text that reports syntactic structure.
+pub trait DemangleWrite {
+    /// Called when we are entering the scope of some AST node.
+    fn push_demangle_node(&mut self, _: DemangleNodeType) {}
+    /// Same as `fmt::Write::write_str`.
+    fn write_string(&mut self, s: &str) -> fmt::Result;
+    /// Called when we are exiting the scope of some AST node for
+    /// which `push_demangle_node` was called.
+    fn pop_demangle_node(&mut self) {}
+}
+
+impl<W: fmt::Write> DemangleWrite for W {
+    fn write_string(&mut self, s: &str) -> fmt::Result {
+        fmt::Write::write_str(self, s)
+    }
+}
+
+impl<'a, T> Symbol<&'a T>
+    where T: AsRef<[u8]> + ?Sized,
+{
     /// Parse a mangled symbol from input and return it and the trailing tail of
     /// bytes that come after the symbol.
     ///
@@ -234,16 +277,16 @@ impl<T> Symbol<T> {
     /// let demangled = sym.to_string();
     /// assert_eq!(demangled, "space::foo(int, bool, char)");
     /// ```
-    pub fn with_tail(input: &[u8]) -> Result<(BorrowedSymbol, &[u8])> {
+    pub fn with_tail(input: &'a T) -> Result<(BorrowedSymbol<'a>, &'a [u8])> {
         let mut substitutions = subs::SubstitutionTable::new();
 
         let ctx = ParseContext::default();
-        let idx_str = IndexStr::new(input);
+        let idx_str = IndexStr::new(input.as_ref());
         let (parsed, tail) = ast::MangledName::parse(&ctx, &mut substitutions, idx_str)?;
         debug_assert!(ctx.recursion_level() == 0);
 
         let symbol = Symbol {
-            raw: input,
+            raw: input.as_ref(),
             substitutions: substitutions,
             parsed: parsed,
         };
