@@ -1,10 +1,11 @@
 //! Types dealing with the substitutions table.
 
+use super::DemangleWrite;
 use ast;
 use std::fmt;
-use std::io;
 use std::iter::FromIterator;
 use std::ops::Deref;
+use vec::Vec;
 
 /// An enumeration of all of the types that can end up in the substitution
 /// table.
@@ -29,13 +30,13 @@ pub enum Substitutable {
 
 impl<'subs, W> ast::Demangle<'subs, W> for Substitutable
 where
-    W: 'subs + io::Write,
+    W: 'subs + DemangleWrite,
 {
     fn demangle<'prev, 'ctx>(
         &'subs self,
         ctx: &'ctx mut ast::DemangleContext<'subs, W>,
         scope: Option<ast::ArgScopeStack<'prev, 'subs>>,
-    ) -> io::Result<()> {
+    ) -> fmt::Result {
         match *self {
             Substitutable::UnscopedTemplateName(ref name) => name.demangle(ctx, scope),
             Substitutable::Type(ref ty) => ty.demangle(ctx, scope),
@@ -51,11 +52,21 @@ impl<'a> ast::GetLeafName<'a> for Substitutable {
         match *self {
             Substitutable::UnscopedTemplateName(ref name) => name.get_leaf_name(subs),
             Substitutable::Prefix(ref prefix) => prefix.get_leaf_name(subs),
+            Substitutable::Type(ref ty) => ty.get_leaf_name(subs),
             _ => None,
         }
     }
 }
 
+impl ast::IsCtorDtorConversion for Substitutable {
+    fn is_ctor_dtor_conversion(&self, subs: &SubstitutionTable) -> bool {
+        match *self {
+            Substitutable::Prefix(ref prefix) => prefix.is_ctor_dtor_conversion(subs),
+            _ => false,
+        }
+    }
+}
+
 /// The table of substitutable components that we have parsed thus far, and for
 /// which there are potential back-references.
 #[doc(hidden)]
@@ -74,6 +85,10 @@ impl fmt::Debug for SubstitutionTable {
         f.pad("SubstitutionTable ")?;
         f.debug_map()
             .entries(self.substitutions.iter().enumerate())
+            .finish()?;
+        f.pad("non_substitutions ")?;
+        f.debug_map()
+            .entries(self.non_substitutions.iter().enumerate())
             .finish()
     }
 }
